from sqlalchemy import select, func


def test_upsert_role():
    assert str(select(func.puddl_upsert_role('unittest', 'unittest'))) == (
        'SELECT puddl_upsert_role(:puddl_upsert_role_2, :puddl_upsert_role_3) '
        'AS puddl_upsert_role_1'
    )



import unittest
from unittest import mock

from puddl import source
from test import PuddlTest


class Test(PuddlTest):
    def test_from_hostname(self):
        with mock.patch("socket.gethostname", lambda: "foohost"):
            assert source.from_hostname("obs") == "obs@foohost"


if __name__ == "__main__":
    unittest.main()



from .datum import Datum


def dump(source, path):
    d = Datum(source=source, path=path)
    return d.to_json()


def get_db():
    from .db import DB

    return DB()


def index(source, path):
    db = get_db()
    d = Datum(source=source, path=path)
    db.stream_datum(source, d)
    return d



#!/usr/bin/env python
import csv
import subprocess
import sys
from datetime import datetime
from pathlib import Path

import pandas as pd
from pydantic import BaseModel

from puddl.felix.git import CommitParseError, log, EmptyLogError, GitLogError

here = Path(__file__).parent


class Record(BaseModel):
    repo_path: str
    hash: str
    dt: datetime
    file_path: str


class Commit:
    def __init__(self, txt: str):
        self._raw = txt
        self.lines = self._raw.strip().split('\n')
        self.header = self.lines[0]
        try:
            h, d, e, s = self.header.split(' ', 3)
        except ValueError:
            raise CommitParseError(f'could not parse header from "{self._raw}"')
        self.hash = h
        self.dt = datetime.fromisoformat(d)
        self.email = e
        self.subject = s
        self.paths = self.lines[1:]

    def __str__(self):
        return f'{self.header} ({len(self.paths)})'


class Repo:
    def __init__(self, path: Path, author=None, since_commit=None):
        self.path = path
        self.author = author
        self.since_commit = since_commit
        self._log = self.get_log()
        if self._log == '':
            raise EmptyLogError(f'{self} is empty for author={self.author}')

    def __str__(self):
        return f'{self.path}'

    def get_log(self):
        cmd = ['git', 'log', '--reverse', "--pretty=format:%H %aI %cE %s", "--stat", "--name-only"]
        if self.author is not None:
            cmd.extend(["--author", self.author])
        if self.since_commit is not None:
            cmd.append(self.since_commit)
        try:
            return subprocess.check_output(cmd, cwd=self.path, encoding='utf-8', stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as e:
            raise GitLogError(f'{self}: {e.stderr}')

    def commits(self) -> [Commit]:
        # records are separated by empty lines
        _records = self._log.strip().split('\n\n')
        for rec in _records:
            if rec is None:
                continue
            yield Commit(rec)

    def __iter__(self):
        try:
            commits = self.commits()
            for commit in commits:
                try:
                    for file_path in commit.paths:
                        yield Record(
                            repo_path=str(self.path),
                            hash=commit.hash,
                            dt=commit.dt,
                            file_path=file_path,
                        )
                except CommitParseError as e:
                    log.warning(f'{self}: {e}')
        except EmptyLogError as e:
            log.warning(e)

    def as_df(self):
        return pd.DataFrame(record.dict() for record in self)

    def as_csv(self, fd):
        fieldnames = list(Record.schema()["properties"].keys())
        writer = csv.DictWriter(fd, fieldnames=fieldnames, quoting=csv.QUOTE_ALL)
        writer.writeheader()
        for record in self:
            writer.writerow(record.dict())


if __name__ == '__main__':
    Repo(Path('.')).as_csv(sys.stdout)



#!/usr/bin/env python
import json

from puddl.pg import DB

app = DB('exif')

data = []
rows = app.engine.execute('SELECT id, lat, lng, alt, thumb, url, dt, since_start FROM markers ORDER BY dt ASC')

for row in rows:
    d = dict(zip(rows.keys(), row))
    data.append(d)

jsdata = json.dumps(data)

tmin = app.engine.execute('SELECT min(dt) FROM markers').scalar()
tmax = app.engine.execute('SELECT max(dt) FROM markers').scalar()
print(f"window.tmin = '{tmin}'")
print(f"window.tmax = '{tmax}'")
print(f"window.exif_data = {jsdata}")



import logging
import os
import subprocess

import click
import psycopg2
import sqlalchemy

from puddl.cli.base import root
from puddl.conf import DBConfig
from puddl.db.tool import connect_to_db
from puddl.pg import DB, list_databases

log = logging.getLogger(__package__)


@root.command()
def init():
    """
    Initializes the database with stuff that puddl depends on, e.g. "FUNCTION puddl_upsert_role".
    """
    import pkg_resources

    conf = DBConfig()
    log.debug(conf.url)
    engine = sqlalchemy.create_engine(conf.url)

    pkg = 'puddl.pg'
    subdir = 'init_sql'
    files = pkg_resources.resource_listdir(pkg, subdir)
    resource_names = ['/'.join((subdir, f)) for f in files]
    for name in resource_names:
        log.debug(f'running file from pkg {pkg}: {name}')
        binary = pkg_resources.resource_string(pkg, name)
        txt = sqlalchemy.text(binary.decode('utf-8'))
        engine.execute(txt)


@root.command()
@click.argument('app', required=False)
def health(app):
    conf = DBConfig(app)
    try:
        connect_to_db(conf.url)
        log.info('database available')
    except psycopg2.OperationalError as e:
        log.debug('', exc_info=e)
        print(e)
        raise SystemExit(1)


@root.command()
def ls():
    conf = DBConfig()
    print('\n'.join(list_databases(conf.url)))


@root.command()
@click.argument('app', required=False)
@click.option('--ignore-errors', default=False, help='Continue even if there are errors')
def shell(app, ignore_errors):
    conf = DBConfig(app)
    conf.PGAPPNAME = 'puddl db shell'
    shell_env = os.environ.copy()
    shell_env.update(dict(conf))
    cmd = ['psql']
    if not ignore_errors:
        cmd.extend(['-v', 'ON_ERROR_STOP=1'])
    # want to see good error handling for this kind of thing?
    # https://github.com/pallets/click/blob/19fdc8509a1b946c088512e0e5e388a8d012f0ce/src/click/_termui_impl.py#L472-L487
    return_code = subprocess.Popen(cmd, env=shell_env).wait()
    raise SystemExit(return_code)


@root.command()
@click.argument('app', required=False)
def url(app):
    """
    print DB URL

    Useful for things like this:

        from sqlalchemy import create_engine
        database_url = 'postgresql://puddl:aey1Ki4oaZohseWod2ri@localhost:13370/puddl'
        engine = create_engine(database_url, echo=False)
        df.to_sql('bp', engine)
    """
    db = DB(app)
    print(db.db_config.url)


@root.command()
@click.argument('app', required=False)
@click.option('-E', '--no-export', is_flag=True, help='do not print export')
def env(app, no_export):
    """
    Prints the database's environment.
    WARNING: This dumps your password. Use it like this:

        source <(puddl db env)
    """
    conf = DBConfig(app)
    for k, v in conf.items():
        if no_export:
            print(f'{k}={v}')
        else:
            print(f'export {k}={v}')


@root.command()
@click.argument('app', required=False)
def pastable(app):
    """
    Print database settings in a way that is suited for copy and pasting.
    """
    conf = DBConfig(app)

    human2slot = {
        'Hostname': 'PGHOST',
        'Port': 'PGPORT',
        'Username': 'PGUSER',
        'Password': 'PGPASSWORD',
        'Database': 'PGDATABASE',
        'App': 'PGAPPNAME',
    }

    records = []
    for name, key in human2slot.items():
        value = conf[key]
        records.append(f'  {name}\n{value}\n')

    print('\n'.join(records))


@root.command()
def sessions():
    """
    Lists active sessions.
    """
    conf = DBConfig()
    conf.PGAPPNAME = 'db sessions'
    query = """SELECT pid AS process_id,
           usename AS username,
           datname AS database_name,
           client_addr AS client_address,
           application_name,
           backend_start,
           state,
           state_change
    FROM pg_stat_activity;"""
    click.echo(subprocess.check_output(['psql', '-c', query], encoding='utf-8', env=conf))


@root.command()
@click.argument('old')
@click.argument('new')
def mv(old, new):
    conf = DBConfig()
    query = f"CALL puddl_rename_schema('{old}', '{new}')"
    subprocess.check_call(['psql', '-c', query], env=conf)


@root.command()
@click.argument('app')
def rm(app):
    conf = DBConfig()
    query = f"DROP DATABASE IF EXISTS {app}"
    subprocess.check_call(['psql', '-c', query], env=conf)


def main():
    root(auto_envvar_prefix='PUDDL')


if __name__ == '__main__':
    main()



import json
import logging
import os

from pathlib import Path

log = logging.getLogger(__name__)

path = Path("~/.puddlrc").expanduser()

_config = {"db": {"drivername": "postgresql"}}


def read():
    global _config
    with path.open() as f:
        _config = json.load(f)


DB2ENV = {
    "username": "PGUSER",
    "password": "PGPASSWORD",
    "host": "PGHOST",
    "port": "PGPORT",
    "database": "PGDATABASE",
}


def from_env():
    global _config
    converted = {a: os.environ[b] for a, b in DB2ENV.items()}
    _config["db"] = {**_config["db"], **converted}


def to_env():
    global _config
    db_env_vars = {b: _config["db"][a] for a, b in DB2ENV.items()}
    return {**os.environ.copy(), **db_env_vars}


def write():
    with path.open("w") as f:
        json.dump(_config, f, sort_keys=True, indent=2)
        f.write("\n")


def get(section):
    read()
    return _config[section]



import datetime
import unittest

from puddl import Datum
from test import PuddlTest


class Test(PuddlTest):
    def test(self):
        x = Datum()
        assert x.uuid is not None
        assert x["uuid"] is not None
        assert x["load_dt"] is not None
        # make sure we can parse the "load_dt"
        datetime.datetime.fromisoformat(x["load_dt"])
        Datum(foo="bar")
        Datum({"bar": "baz"})

    def test_update(self):
        d = Datum()
        d.update({"foo": "bar"}, x=23)
        assert d["foo"] == "bar"
        assert d["x"] == 23
        # can update with None
        d.update(None)
        assert d["foo"] == "bar"
        assert d["x"] == 23

    def test_file_stuff(self):
        d = Datum(path=self.data1)
        assert d["path"] == str(self.data1)
        assert d.path == self.data1
        assert "filetype" not in d, "magic has no info about this simple text file"
        assert "stat" in d
        s = d.to_json()
        self.assertTrue(s)

    def test_ffprobe(self):
        d = Datum(path=self.mkv)
        assert d["ffprobe"]["format"]["start_time"] == "0.000000"
        assert d["ffprobe"]["format"]["duration"] == "4.934000"
        assert d["ffprobe"]["format"]["format_name"] == "matroska,webm"


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import typing as t
from pathlib import Path


class Video:
    def __init__(self, data: bytes, encoding: str):
        self.data = data
        self.encoding = encoding

    @classmethod
    def from_path(cls, path: Path) -> Video:
        raise NotImplementedError


class Thumb:
    def __init__(self, data: bytes):
        self.data = data

    def to_png(self) -> t.IO[bytes]:
        raise NotImplementedError


def make_thumbs(video: Video) -> [Thumb]:
    """
    :param video: video Bytewurst
    """
    raise NotImplementedError


if __name__ == '__main__':
    import sys

    video = Video.from_path(sys.argv[1])
    thumbs = make_thumbs(video)
    for i, thumb in enumerate(thumbs):
        png = thumb.to_png()
        path = Path(f'thumbnail_{i:03}.png')
        path.write_bytes(png)



import unittest
from pathlib import Path


class PuddlTest(unittest.TestCase):
    DATADIR = Path(__file__).parent / "data"

    def setUp(self) -> None:
        self.data1 = self.DATADIR / "data1.json"
        self.data2 = self.DATADIR / "data2"
        self.mkv = self.DATADIR / "2019-11-01_18-26-42.mkv"



import logging
import re
import subprocess

import psycopg2.errorcodes
import sqlalchemy
from sqlalchemy import inspect
from sqlalchemy.engine import Engine
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import sessionmaker

from puddl.conf import DBConfig
from puddl.exc import DBException

log = logging.getLogger(__package__)


class DB:
    """
    >>> db = DB('test')

    This creates a Postgres ROLE and SCHEMA with the name "sunrise".

    Please instantiate this class once per process, because it uses SQLAlchemy's engine,
    thus similar rules apply. "[...] the Engine is most efficient when created just once
    at the module level of an application, not per-object or per-function call"
    -- https://docs.sqlalchemy.org/en/13/core/connections.html#basic-usage

    An app holds DB configuration and an engine:

    >>> db.db_config
    DBConfig('sunrise')
    >>> db.engine
    Engine(postgresql://sunrise:***@localhost:13370/puddl?application_name=sunrise)
    """

    name: str
    db_config: DBConfig
    engine: Engine

    def __init__(self, name):
        log.debug(f'initializing DB({repr(name)})')
        self.name = name

        if self.name is None:
            self.db_config = DBConfig()
        else:
            # https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
            if re.match(r'[_a-z][a-z0-9]*', name) is None:
                raise DBException('Invalid DB name')
            # our config is solely based on "name"
            self.db_config = DBConfig(name)
            self._init_named_database(self.db_config)

        self.engine = sqlalchemy.create_engine(self.db_config.url)
        session_cls = sessionmaker(self.engine)
        self.session = session_cls()
        self.autocommit_engine = self.engine.execution_options(isolation_level="AUTOCOMMIT")

    @staticmethod
    def _init_named_database(db_config: DBConfig):
        x = db_config
        # get global credentials (to upsert the app itself)
        root_db_config = DBConfig()
        # CREATE DATABASE only works in Postgres' AUTOCOMMIT mode
        root_engine = sqlalchemy.create_engine(root_db_config.url, isolation_level="AUTOCOMMIT")
        # https://docs.sqlalchemy.org/en/14/core/connections.html#using-transactions
        with root_engine.connect() as root_conn:
            log.debug(f'creating role {x.PGUSER}')
            root_conn.execute(sqlalchemy.func.puddl_upsert_role(x.PGUSER, x.PGPASSWORD))
            log.debug(f'the role "puddl" is part of the role {x.PGUSER}')
            root_conn.execute(f"GRANT {x.PGUSER} TO puddl")
            try:
                log.debug(f'creating database {x.PGDATABASE}')
                root_conn.execute(f"CREATE DATABASE {x.PGDATABASE} WITH OWNER {x.PGUSER}")
            except ProgrammingError as e:
                if e.orig.pgcode == psycopg2.errorcodes.DUPLICATE_DATABASE:
                    log.debug(f'database {x.PGDATABASE} already exists')
                    pass
                else:
                    raise

    def has_table(self, table_name):
        return inspect(self.engine).has_table(table_name)

    def df_dump(self, df, table_name, index=True, drop_cascade=False):
        """
        Dumps a DataFrame to table_name.
        WARNING! This replaces the table if it exists.
        """
        if drop_cascade:
            self.engine.execute(f'DROP TABLE IF EXISTS {table_name} CASCADE')
        if index:
            # follow SQL conventions for the index
            df.index += 1
        return df.to_sql(table_name, self.engine, if_exists='replace', index=index)

    def df_append(self, df, table_name: str, index=True):
        if index:
            # follow SQL conventions for the index
            df.index += 1
        return df.to_sql(table_name, self.engine, if_exists='append', index=index)

    def psql(self):
        db_env = self.db_config
        subprocess.Popen('psql', env=db_env).wait()



#!/usr/bin/env python

import sys
from pathlib import Path

import pandas as pd

from puddl.db.alchemy import App

fd = Path(sys.argv[1]).expanduser().open()
data_iter = (line.rstrip("\n") for line in fd)
df = pd.DataFrame(data_iter, columns=["line"])

app = App("bash_profile")
app.df_dump(df, "lines", drop_cascade=True)



from datetime import datetime
import streamlit as st
import pandas as pd

from puddl.db.alchemy import App

# https://deckgl.readthedocs.io/en/latest/
import pydeck as pdk
from pydeck.data_utils.viewport_helpers import compute_view

st.set_page_config(layout='wide')
app = App('exif')


def select2datetime(s: str):
    dt0_iso = app.engine.execute(s).scalar()
    return datetime.fromisoformat(dt0_iso)


dt_min = select2datetime('SELECT min(dt) as dt FROM markers')
dt_max = select2datetime('SELECT max(dt) as dt FROM markers')
# st.sidebar.text(f'dt_min={dt_min}')
# st.sidebar.text(f'dt_max={dt_max}')


dt_range = st.sidebar.date_input('select date range', [dt_min, dt_max], min_value=dt_min, max_value=dt_max)


def load_data():
    ts0 = datetime.combine(dt_range[1], datetime.min.time())
    ts1 = datetime.combine(dt_range[1], datetime.max.time())
    data = pd.read_sql(
        'SELECT * FROM markers WHERE dt_ts BETWEEN %(ts0)s AND %(ts1)s',
        app.engine,
        params={'ts0': ts0, 'ts1': ts1},
    )

    def lowercase(x):
        return str(x).lower()

    data.rename(lowercase, axis='columns', inplace=True)
    return data


data_load_state = st.sidebar.text('Loading data...')
data = load_data()
data_load_state.text("Done loading.")


def thumb2icondata(t):
    return {
        # Icon from Wikimedia, used the Creative Commons Attribution-Share Alike 3.0
        # Unported, 2.5 Generic, 2.0 Generic and 1.0 Generic licenses
        "url": t,
        "width": 242,
        "height": 242,
        "anchorY": 242,
    }


data['icon_data'] = data['thumb'].map(thumb2icondata)

# Note that there is no independent scrolling (vertical overflow for a colum with height) for columns.
# https://github.com/streamlit/streamlit/issues/2447
col1, col2 = st.columns(2)

with col1:
    layer = pdk.Layer(
        "IconLayer",
        data,
        get_position="[lng, lat]",
        get_icon="icon_data",
        get_size=4,
        size_scale=15,
        pickable=True,
    )
    # Set the viewport location
    view_state = compute_view(data[['lng', 'lat']])
    view_state.pitch = 40.5
    view_state.bearing = -27.36

    r = pdk.Deck(
        map_style="mapbox://styles/mapbox/light-v9",
        layers=[layer],
        initial_view_state=view_state,
        tooltip={"html": '<img src="{thumb}">', "style": {"color": "white"}},
    )
    st.pydeck_chart(r)

    if st.sidebar.checkbox('show raw data'):
        st.subheader('Raw data')
        st.write(data)

with col2:
    with st.container():
        for t in data['thumb']:
            st.image(t)



import sqlalchemy

from puddl.typing import URL


def list_databases(url: URL):
    engine = sqlalchemy.create_engine(url)
    return [row[0] for row in engine.execute('SELECT name FROM puddl_databases').fetchall()]


def list_schemas(url: URL):
    engine = sqlalchemy.create_engine(url)
    inspection_result = sqlalchemy.inspect(engine)
    return inspection_result.get_schema_names()



import altair as alt
import pandas as pd

data = pd.DataFrame({'x': ['A', 'B', 'C', 'D', 'E'], 'y': [5, 3, 6, 7, 2]})
chart = (
    alt.Chart(data)
    .mark_bar()
    .encode(
        x='x',
        y='y',
    )
    .interactive()
)

chart.show()



def test_empty_env():
    import os
    from puddl.db.alchemy import DBConfig
    for k in DBConfig.__slots__:
        if k in os.environ:
            del os.environ[k]
    x = DBConfig('sunrise')
    assert x is not None



import collections.abc
import json
import logging
import os
from pathlib import Path
from urllib.parse import urlencode

from puddl.typing import URL

log = logging.getLogger(__package__)


class ConfigError(Exception):
    pass


class PuddlRC:
    PATH = Path.home() / '.puddlrc'

    @classmethod
    def exists(cls):
        return cls.PATH.exists()

    @classmethod
    def read(cls):
        with cls.PATH.open() as f:
            try:
                return json.load(f)
            except json.decoder.JSONDecodeError as e:
                raise ConfigError(
                    f'{cls.PATH}:{e.lineno} column {e.colno} char {e.pos}: {e.msg}'
                )

    @classmethod
    def write(cls, data):
        x = json.dumps(data, indent=2, sort_keys=True)
        with cls.PATH.open('w') as f:
            f.write(x)


class DBConfig(collections.abc.Mapping):
    """
    Abstracts a subset of the libpq environment variables defined by
    https://www.postgresql.org/docs/current/libpq-envars.html

    >>> puddl_root_db_config = DBConfig()
    >>> puddl_root_db_config.url
    'postgresql://puddl:Eexoozohgh4ni5deimoh@127.0.0.1:13370/puddl?application_name=puddl+config+init'

    Given an app name, user and password are set accordingly

    >>> app_db_config = DBConfig('foo')
    >>> app_db_config.url
    'postgresql://foo:foo@127.0.0.1:13370/foo?application_name=foo'
    """

    __slots__ = ['PGUSER', 'PGPASSWORD', 'PGHOST', 'PGPORT', 'PGDATABASE', 'PGAPPNAME']

    def __init__(self, name=None):
        conf_vars_missing = []
        conf = PuddlRC.read()
        for key in self.__slots__:
            try:
                value = conf[key]
                setattr(self, key, value)
            except KeyError:
                conf_vars_missing.append(key)
        if conf_vars_missing:
            raise ConfigError(f'Missing variables: {conf_vars_missing}. Please initialize {PuddlRC.PATH}')

        if name is not None:
            self.PGDATABASE = name
            self.PGUSER = name
            self.PGPASSWORD = name
            self.PGAPPNAME = name

    @property
    def url(self) -> URL:
        db_params = {
            'application_name': self.PGAPPNAME
        }
        params = urlencode(db_params)
        return f'postgresql://{self.PGUSER}:{self.PGPASSWORD}' \
               f'@{self.PGHOST}:{self.PGPORT}' \
               f'/{self.PGDATABASE}?{params}'

    def __str__(self):
        return self.PGUSER

    def __repr__(self):
        return f"DBConfig('{self.PGUSER}')"

    # implement Mapping ABC
    def keys(self):
        return self.__slots__

    def __getitem__(self, item):
        if item not in self.__slots__:
            raise KeyError
        return getattr(self, item)

    def __len__(self):
        return len(self.__slots__)

    def __iter__(self):
        return iter(self.__slots__)



#!/usr/bin/env python
import mailbox
import re
import sys


p = sys.argv[1]
d = mailbox.Maildir(p)

RE_VALID = re.compile(r'([a-z0-9._-]+)@felixhummel.de')

broken_headers = []

for message in d:
    to = message['to']
    if to is None:
        continue
    try:
        m = RE_VALID.search(to)
    except TypeError:
        broken_headers.append(message)
    if m is not None:
        print(m.group(1))
        pass

# one could handle broken_headers here ^^



from pathlib import Path

from .load import img2dict


def test_img2dict():
    example_jpg = Path(__file__).parent / 'example.jpg'
    d = img2dict(example_jpg)
    assert d['_name'] == 'example.jpg'
    assert d['_path'].endswith('example.jpg')
    assert d['thumb'].startswith('data:image')
    assert d['GPS GPSLatitude'] == '[47, 40, 50492639/1000000]'
    assert d['GPS GPSLongitude'] == '[12, 8, 12246359/1000000]'
    assert d['Image GPSInfo'] == '672'



import pytest

from puddl.exc import DBException
from puddl.pg import DB


def test_create_and_connect():
    db1 = DB('unittest')
    db1.engine.execute('SELECT 1').first()

    db2 = DB('unittest')
    db2.engine.execute('SELECT 1').first()


def test_puddl_access():
    from puddl.conf import DBConfig
    import sqlalchemy
    db = DB('unittest')
    with db.engine.connect() as conn:
        conn.execute('CREATE TABLE IF NOT EXISTS test_puddl_access (x INTEGER)')
        conn.execute('DELETE FROM test_puddl_access')
        conn.execute('INSERT INTO test_puddl_access (x) VALUES (23)')
    # puddl may read and modify data
    # stuff should still be owned by the application
    root_db_config = DBConfig()
    root_db_config.PGDATABASE = db.name
    root_engine = sqlalchemy.create_engine(root_db_config.url)
    with root_engine.connect() as conn:
        result = conn.execute('SELECT * FROM test_puddl_access')
        assert result.one()._mapping == {'x': 23}


def test_invalid_name():
    with pytest.raises(DBException):
        DB('`')


def test_df_dump():
    import logging
    logging.getLogger('puddl.db.alchemy').setLevel(logging.DEBUG)
    import pandas as pd
    db = DB('unittest')
    df = pd.DataFrame({'foo': [1, 2], 'bar': [5, 6]})
    db.df_dump(df, 'test_df_dump')



import subprocess
from itertools import chain

import sqlalchemy
from sqlalchemy import inspect
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import CreateColumn
from sqlalchemy.sql.functions import GenericFunction
from sqlalchemy.types import Text

from puddl.conf import DBConfig

# https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#postgresql-10-identity-columns
from puddl.typing import URL


@compiles(CreateColumn, 'postgresql')
def use_identity(element, compiler, **kw):
    text = compiler.visit_create_column(element, **kw)
    text = text.replace("SERIAL", "INT GENERATED BY DEFAULT AS IDENTITY")
    return text


class Upserter:
    """
    Finds all columns with a unique constraint and upserts based on those.
    """

    def __init__(self, session, model):
        self.session = session
        self.model = model
        inspector = Inspector.from_engine(session.bind)
        _xs = inspector.get_unique_constraints(self.model.__tablename__, schema=self.model.__table__.schema)
        list_of_lists = [x['column_names'] for x in _xs]
        self.unique_columns = set(chain.from_iterable(list_of_lists))

    def upsert(self, **data):
        # read this carefully:
        # https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#sqlalchemy.dialects.postgresql.dml.Insert.on_conflict_do_update
        columns = set(data.keys())
        update_columns = columns - self.unique_columns
        insert_stmt = insert(self.model).values(**data)
        excluded_mapping = {  # LHS: column name, RHS: Column instance
            col: getattr(insert_stmt.excluded, col) for col in update_columns
        }
        do_update_stmt = insert_stmt.on_conflict_do_update(
            index_elements=self.unique_columns,
            set_=excluded_mapping,
        )
        self.session.execute(do_update_stmt)


# https://docs.sqlalchemy.org/en/14/core/functions.html#sqlalchemy.sql.functions.GenericFunction
# noinspection PyPep8Naming
class puddl_upsert_role(GenericFunction):
    type = Text





import logging
import socket

import click
import pendulum

import puddl
import puddl.config

log = logging.getLogger(__name__)
LOG_LEVELS = ["CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG"]


@click.group()
@click.option("--debug/--no-debug")
@click.option(
    "--log-level", type=click.Choice(LOG_LEVELS, case_sensitive=False), default="INFO"
)
def root(debug, log_level):
    if debug:
        logging.basicConfig(level="DEBUG")
        log.debug(f'debug mode enabled (run "PUDDL_DEBUG=false" to disable)')
    else:
        logging.basicConfig(level=log_level)
    puddl.config.read()


@root.command()
def date():
    dt = pendulum.now()
    print(dt.format("YYYY-MM-DD__HH-mm-ss__ZZ"))


_default_source = socket.gethostname()


@root.command()
@click.option("--db/--no-db", default=True)
def setup(db):
    puddl.config.from_env()
    puddl.config.write()
    if db:
        puddl.get_db().migrate()


@root.command()
@click.option("-s", "--source")
@click.option("--dump", default=False)
@click.argument("file", type=click.Path(exists=True), nargs=-1)
def index(source, dump, file):
    for path in file:
        if dump:
            print(puddl.dump(source, path))
        else:
            # an "infer t0" would be nice here...
            puddl.index(source, path)
    print(f"indexed {len(file)} files")


@root.command()
@click.argument("stream")
@click.argument("comment")
def comment(stream, comment):
    print(stream, comment)


@root.group()
def ingress():
    pass


@ingress.command(deprecated=True)
@click.option("--binary")
def stdin(binary):
    if binary:
        raise NotImplementedError
    db = puddl.get_db()
    source = _default_source
    db.stream(source=source, ingress="stdin", in_stream=click.get_text_stream("stdin"))


@root.group(name="db")
def database():
    pass


@database.command()
def status():
    if not puddl.get_db().is_ok():
        click.get_current_context().exit(code=1)


@database.command()
def env():
    for k, v in puddl.config.to_env().items():
        if k.startswith("PG"):
            print(f"export {k}={v}")


@root.group()
def dev():
    pass


@dev.command()
def drop_all_tables():
    puddl.get_db().drop_all_tables()


def main():
    # Note that this function is referenced in `setup.py`.
    root(auto_envvar_prefix="PUDDL")


if __name__ == "__main__":
    # https://click.palletsprojects.com/en/7.x/options/#values-from-environment-variables
    main()



import logging
import os

import click

import puddl

LOG_LEVELS = ['CRITICAL', 'FATAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG']


@click.group()
@click.option(
    '-l', '--log-level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=os.environ.get('LOGLEVEL')
)
@click.option('-d', '--debug', is_flag=True, help='sets log level to DEBUG. ignores "--log-level"')
@click.option('--loggers', default='', help='comma-separated logger names')
@click.version_option(version=puddl.__version__)
def root(log_level, debug, loggers):
    if debug:
        log_level = 'DEBUG'
    # longest level is WARNING with 7 character
    logging.basicConfig(level=log_level, format='%(name)s %(levelname)-7s %(msg)s')
    if loggers:
        for name in loggers.split(','):
            logging.getLogger(name).setLevel(log_level)
    pass



import pytest
from puddl.pg import DB


def test_foo():
    import psycopg2.errorcodes
    from sqlalchemy.exc import ProgrammingError

    db = DB('psycoerror')

    with pytest.raises(ProgrammingError) as e:
        db.engine.execute('CREATE TABLE foo (x INT)')
        db.engine.execute('CREATE TABLE foo (x INT)')
        assert e.orig.pgcode == psycopg2.errorcodes.DUPLICATE_TABLE



import logging

log = logging.getLogger(__name__)


class GitError(Exception):
    pass


class GitLogError(GitError):
    pass


class EmptyLogError(GitError):
    pass


class CommitParseError(GitError):
    pass



import json
import logging
from collections import UserDict
from pathlib import Path
from uuid import uuid4

import ffmpeg
import filetype
import pendulum

log = logging.getLogger(__name__)


class Datum(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.uuid = self["uuid"] = str(uuid4())
        self.load_dt = pendulum.now()
        self["load_dt"] = self.load_dt.to_iso8601_string()
        # special case: files haves paths. we have lots of files.
        if "path" in self:
            self.path = Path(self.pop("path"))
            self["path"] = str(self.path)
            self["filename"] = self.path.name
            self["uri"] = self.path.absolute().as_uri()
            self["filetype"] = self._guess_filetype()
            self["stat"] = self._stat()
            # TODO EAFP instead of LBYL
            if self._is_video():
                self["ffprobe"] = self._ffprobe()

    # ignore when None is passed as value
    def update(self, *args, **kwargs):
        if len(args) == 1 and args[0] is None:
            return
        else:
            super().update(*args)
        super().update(**kwargs)

    def __setitem__(self, key, value):
        if value is None:
            log.debug(f"{key} is None")
            return
        super().__setitem__(key, value)

    # -----------------------------------

    # file handling (maybe refac this)
    def _guess_filetype(self):
        x = filetype.guess(str(self.path))
        if x is not None:
            return {"extension": x.extension, "mime": x.mime}

    def _stat(self):
        s = self.path.stat()
        attrs = [a for a in dir(s) if a.startswith("st_")]
        return {a: getattr(s, a) for a in attrs}

    def _ffprobe(self):
        return ffmpeg.probe(self.path)

    def _is_video(self):
        try:
            return self["filetype"]["mime"].startswith("video/")
        except KeyError:
            return False

    # -------------

    def to_json(self):
        return json.dumps(self.data, sort_keys=True, indent=2)



import logging
from pathlib import Path

import pandas as pd

from puddl import App
from puddl.felix.git import GitError
from puddl.felix.git.repo2rows import Repo

log = logging.getLogger(__package__)


class GitApp(App):
    def __init__(self, name='git'):
        super().__init__(name)

    def truncate(self):
        self.engine.execute('TRUNCATE raw')

    def load(self, paths: [Path], author=None, since_commit=None):
        for path in paths:
            try:
                repo = Repo(path, author, since_commit)
                df = repo.as_df()
                if df.empty:
                    continue
                df['dt'] = pd.to_datetime(df['dt'], utc=True).dt.tz_convert('Europe/Berlin')
                self.df_append(df, 'raw')
                log.info(f'{path} with {len(df)} records')
            except GitError:
                pass



import logging

from sqlalchemy import create_engine, Column, Integer, Text, DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from puddl import Datum

log = logging.getLogger(__name__)

Base = declarative_base()


class Stream(Base):
    __tablename__ = "stream"
    id = Column(Integer, primary_key=True)
    source = Column(Text, index=True, nullable=False)  # e.g. "f@obs"
    meta = Column(JSONB, nullable=False)
    created_dt = Column(DateTime(timezone=True), server_default=func.now())

    def __str__(self):
        return f"{self.source}: {self.best_guess_timestamp}"

    @property
    def best_guess_timestamp(self):
        # TODO infer from meta
        return self.created_dt


class DB:
    def __init__(self):
        from puddl import config

        db_args = config.get("db")
        self.url = URL(**db_args)
        self.engine = create_engine(self.url)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()

    def drop_all_tables(self):
        Base.metadata.drop_all(self.engine)

    def is_ok(self):
        try:
            self.engine.execute("SELECT 1")
            log.info("Connection OK")
            return True
        except OperationalError as e:
            log.error(str(e))
            return False

    def migrate(self):
        # Alembic!
        # for now: create stuff
        Base.metadata.create_all(self.engine)

    def stream_simple(self, source, meta):
        stream = Stream(
            source=source,
            meta=meta,
        )
        self.session.add(stream)
        self.session.commit()

    def stream_datum(self, source, d: Datum):
        stream = Stream(
            source=source,
            meta=d.data,
        )
        self.session.add(stream)
        self.session.commit()



#!/usr/bin/env python
import sys
from pathlib import Path

import click

from puddl.cli.base import root
from .app import GitApp
from .repo2rows import Repo
from .utils import list_repo_paths


@root.command()
@click.argument('path', type=str)
def csv(path: str):
    repo = Repo(Path(path).expanduser().absolute())
    print(repo.as_csv(sys.stdout))


# noinspection PyShadowingNames
@root.command()
@click.argument('path', type=str, nargs=-1)
@click.option('-r', '--recursive', is_flag=True, help="Find repos recursively (by looking for .git) and index them")
@click.option('--truncate', is_flag=True, help="Truncate the table 'raw' before loading?")
@click.option('--author', help="Filter by author. See `git log --help`.")
@click.option('--since-commit', help="Only index in the given <revision range>. Check 'git help log'")
def index(path: str, recursive: bool, truncate: bool, author, since_commit):
    """
    Example call:

    puddl-felix-git -linfo index --truncate -r --since-commit HEAD ~
    """
    root_dirs = [Path(d).expanduser().absolute() for d in path]
    app = GitApp()
    if truncate:
        app.truncate()
    if recursive:
        paths = []
        for root_dir in root_dirs:
            paths.extend(list_repo_paths(root_dir))
    else:
        paths = root_dirs
    app.load(paths, author, since_commit)


@root.command()
def truncate():
    app = GitApp()
    app.truncate()


@root.command()
def ls():
    app = GitApp()
    import pandas as pd

    q = "SELECT DISTINCT repo_path || '/' || file_path FROM raw"
    df = pd.read_sql_query(q, app.engine)
    print(df.to_csv(index=False))


def main():
    root(auto_envvar_prefix='PUDDL_FELIX_GIT')


if __name__ == '__main__':
    main()



import unittest

import puddl.config
from test import PuddlTest


class TestConfig(PuddlTest):
    def test(self):
        puddl.config.read()
        puddl.config.to_env()


if __name__ == "__main__":
    unittest.main()



import logging
import os

import click

from puddl.cli.base import root
from puddl.conf import DBConfig, PuddlRC

log = logging.getLogger(__name__)


@root.command()
@click.option('-f', '--force', is_flag=True)
def init(force):
    if PuddlRC.exists():
        if not force:
            raise click.ClickException(f'{PuddlRC.PATH} already exists')
        else:
            log.info(f'forcing overwrite of {PuddlRC.PATH}')
    from dotenv import load_dotenv

    load_dotenv()
    os.environ['PGAPPNAME'] = 'puddl-config init'

    cfg = DBConfig()
    PuddlRC.write(dict(cfg))


@root.command()
def show():
    for k, v in DBConfig().items():
        print(f'{k}={v}')


def main():
    root(auto_envvar_prefix='PUDDL')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import base64
import logging
import os
import sys
from pathlib import Path

import exifread
import pandas as pd

from puddl.pg import DB


def img2dict(path: Path):
    # basics
    result = {'_path': str(path), '_name': path.name}
    # exif
    with path.open('rb') as f:
        exif_data = exifread.process_file(f)
        for k, v in exif_data.items():
            if isinstance(v, exifread.classes.IfdTag):
                exif_data[k] = v.printable
        result.update(exif_data)
    # thumb
    if 'JPEGThumbnail' in result:
        # noinspection PyTypeChecker
        payload = base64.b64encode(result['JPEGThumbnail']).decode('utf-8')
        img = 'data:image/jpeg;base64,' + payload
    else:
        img = (
            'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQ'
            'AAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII='
        )
    result['thumb'] = img
    return result


def main():
    log = logging.getLogger()
    logging.basicConfig(level=os.environ.get('LOGLEVEL', 'WARNING').upper())

    directory = Path(sys.argv[1])

    imgs = list(directory.glob('*.jpg'))
    if not imgs:
        raise Exception('No images')
    log.info(f'processing {len(imgs)} images')

    db = DB('exif')
    log.info(db.engine)

    df = pd.DataFrame(img2dict(img) for img in imgs)
    db.df_dump(df, 's7', drop_cascade=True)


if __name__ == '__main__':
    main()



from puddl.typing import URL


def connect_to_db(url: URL):
    """
    throws psycopg2.OperationalError on failure
    """
    import psycopg2
    connection = psycopg2.connect(url)
    with connection.cursor() as cursor:
        cursor.execute('SELECT 1')



import unittest

import puddl.config
from puddl import Datum
from test import PuddlTest


class TestDB(PuddlTest):
    def test_stream(self):
        db = puddl.get_db()
        db.stream_simple("obs@f", {"test": True})

    def test_stream_datum(self):
        db = puddl.get_db()
        datum = Datum(path=self.mkv)
        db.stream_datum("obs@f", datum)


if __name__ == "__main__":
    unittest.main()



# coding by stackoverflow :>
# https://stackoverflow.com/a/53503693/241240
import os
from pathlib import Path

DOT_GIT = '.git'


def list_repo_paths(root: Path) -> [Path]:
    """
    Given a root directory, find all of its subdirectories that contain a `.git` file.
    Here `.git` can be a directory (for "normal" repos) or a file (for submodules).
    """
    for path, dirs, files in os.walk(root):
        if DOT_GIT in dirs + files:
            yield Path(path)



from pathlib import Path

from flask import Flask, send_from_directory

from puddl.pg import DB

app = Flask(__name__)
pdl = DB('exif')
root = Path('.').absolute()


@app.route('/')
def index():
    return send_from_directory(root, 'index.html')


@app.route('/<path:path>')
def _home(path):
    print(path)
    return send_from_directory(root, path)


@app.route("/space")
def space():
    data = []
    rows = pdl.engine.execute('SELECT id, lat, lng, alt, thumb, url, dt, since_start FROM markers ORDER BY dt ASC')

    for row in rows:
        d = dict(zip(rows.keys(), row))
        data.append(d)
    return {'rows': data}


@app.route("/time")
def time():
    return {
        'tmin': pdl.engine.execute('SELECT min(dt) FROM markers').scalar(),
        'tmax': pdl.engine.execute('SELECT max(dt) FROM markers').scalar(),
    }



import logging
from pathlib import Path

from puddl.felix.git.repo2rows import iter_records


def test_empty_header(caplog):
    caplog.set_level(logging.INFO, logger='puddl.felix.git')
    print(caplog.records)
    records = list(iter_records(Path('/home/felix/github/kimai-in-docker')))
    assert len(records) > 0
